import unittest

from decentnet.modules.cryptography.symetric import AESCipher


class TestAESCipher(unittest.TestCase):

    def test_encrypt_decrypt_with_different_key_sizes(self):
        for key_size in [128, 192, 256]:
            with self.subTest(key_size=key_size):
                password = "testpassword".encode()
                plaintext = b"This is a test message."
                cipher = AESCipher(password=password, key_size=key_size)

                # Encrypt the plaintext
                encrypted_text = cipher.encrypt(plaintext)

                # Decrypt the encrypted text
                decrypted_text = cipher.decrypt(encrypted_text)

                # Assert that the decrypted text is the same as the original plaintext
                self.assertEqual(decrypted_text, plaintext)

    def test_invalid_key_size(self):
        with self.assertRaises(ValueError):
            AESCipher(b"array", 64)  # An unsupported key size


if __name__ == '__main__':
    unittest.main()



import logging

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


def derive_key(salt: bytes, password: bytes, iterations: int, key_length: int) -> bytes:
    """
    Derives a key from the provided salt and password using PBKDF2.

    Args:
        salt: The salt value used in key derivation.
        password: The password to derive the key from.
        iterations: The number of iterations for the key derivation algorithm.
        key_length: The desired length of the derived key in bytes.

    Returns:
        The derived key.

    """
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=key_length,
        salt=salt,
        iterations=iterations,
        backend=default_backend()
    )
    key = kdf.derive(password)
    return key


def encrypt_data(key: bytes, nonce: bytes, plaintext: bytes, aad: bytes) -> bytes:
    """
    Encrypts the provided plaintext using ChaCha20-Poly1305.

    Args:
        key: The encryption key.
        nonce: The nonce value for the encryption.
        plaintext: The data to be encrypted.
        aad: Additional authenticated data.

    Returns:
        The ciphertext.

    """
    cipher = Cipher(algorithms.ChaCha20(key, nonce), mode=None, backend=default_backend())
    encryptor = cipher.encryptor()
    encryptor.authenticate_additional_data(aad)
    ciphertext = encryptor.update(plaintext) + encryptor.finalize()
    return ciphertext


def decrypt_data(key: bytes, nonce: bytes, ciphertext: bytes, aad: bytes) -> bytes:
    """
    Decrypts the provided ciphertext using ChaCha20-Poly1305.

    Args:
        key: The decryption key.
        nonce: The nonce value for the decryption.
        ciphertext: The data to be decrypted.
        aad: Additional authenticated data.

    Returns:
        The decrypted plaintext.

    """
    cipher = Cipher(algorithms.ChaCha20(key, nonce), mode=None, backend=default_backend())
    decryptor = cipher.decryptor()
    decryptor.authenticate_additional_data(aad)
    plaintext = decryptor.update(ciphertext) + decryptor.finalize()
    return plaintext



import unittest

from decentnet.modules.pow.difficulty import Difficulty


class TestDifficultyConversion(unittest.TestCase):
    def test_bytes_conversion_round_trip(self):
        test_cases = [
            (1, 16, 2, 255, 32),
            (10, 1024, 20, 8, 64),
            (0, 32, 1, 1, 1),
            (255, 2048, 255, 255, 255)
        ]

        for t_cost, m_cost, p_cost, n_bits, hash_len_chars in test_cases:
            with self.subTest(t_cost=t_cost, m_cost=m_cost, p_cost=p_cost, n_bits=n_bits,
                              hash_len_chars=hash_len_chars):
                original = Difficulty(t_cost, m_cost, p_cost, n_bits, hash_len_chars)
                converted_bytes = original.to_bytes()
                reconstructed = Difficulty.from_bytes(converted_bytes)

                self.assertEqual(original, reconstructed)


if __name__ == '__main__':
    unittest.main()



from decentnet.consensus.byte_conversion_constants import ENDIAN_TYPE


def int_to_bytes(num: int, buffer=None):
    if num == 0:
        return b'\x00'

        # Calculate the number of bytes needed to represent the integer
    byte_length = (num.bit_length() + 7) // 8

    # If no buffer is provided or it's too small, return the bytes directly
    if buffer is None or len(buffer) < byte_length:
        return num.to_bytes(byte_length, ENDIAN_TYPE)

    # Convert integer to bytes directly into the provided buffer
    num_bytes = num.to_bytes(byte_length, ENDIAN_TYPE)

    buffer[:byte_length] = num_bytes
    return bytes(buffer[:byte_length])



"""Block Byte allocation for serialization and deserialization"""
from decentnet.__version__ import network_version

# This blocks maximum size is intentionally smaller
# to prevent DDOS attacks and ensure decentralized sending
MAXIMUM_BLOCK_SIZE = 4194304  # Bytes
BLOCK_PREFIX_LENGTH_BYTES = 3  # Bytes
INDEX_SIZE = 8  # Bytes
VERSION_SIZE = len(network_version)
NONCE_SIZE = 16  # Bytes
TIMESTAMP_SIZE = 8  # Bytes

MERGED_DIFFICULTY_BYTE_LEN = 13  # Bytes
HASH_LEN = 32  # Bytes
HASH_LEN_BLOCK = 48  # Bytes

RESERVE_BYTE_LEN = 77  # Bytes

MAXIMUM_DATA_SIZE = (MAXIMUM_BLOCK_SIZE - BLOCK_PREFIX_LENGTH_BYTES - INDEX_SIZE -
                     VERSION_SIZE - NONCE_SIZE - TIMESTAMP_SIZE -
                     MERGED_DIFFICULTY_BYTE_LEN - HASH_LEN -
                     HASH_LEN_BLOCK - RESERVE_BYTE_LEN)



import asyncio
from typing import Optional, Tuple

from sqlalchemy import select

from decentnet.modules.db.base import session_scope
from decentnet.modules.db.models import OwnedKeys


class AliasResolver:
    """
    AliasResolver is responsible for resolving the public key and key ID based on a given alias.
    """

    @classmethod
    def get_key_by_alias(cls, alias: str) -> Optional[Tuple[str, int]]:
        return asyncio.run(cls.__get_key_by_alias(alias))

    @classmethod
    async def __get_key_by_alias(cls, alias: str) -> Optional[Tuple[str, int]]:
        """
        Retrieves the public key and key ID based on the given alias.

        Args:
            alias (str): The alias to search for

        Returns:
            Optional[Tuple[str, int]]: A tuple containing the public key and key ID if found, otherwise None.
        """
        async with session_scope() as session:
            # Perform the asynchronous query
            result = await session.execute(
                select(OwnedKeys).where(OwnedKeys.alias == alias)
            )
            key = result.scalar_one_or_none()  # Fetch the first result or None

            if key:
                return key.public_key, key.id

            return None



import asyncio
import logging
import os
import threading
from typing import Optional

from decentnet.consensus.block_sizing import BLOCK_PREFIX_LENGTH_BYTES
from decentnet.consensus.byte_conversion_constants import ENDIAN_TYPE
from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.logger.log import setup_logger
from decentnet.modules.tcp.db_functions import remove_alive_beam_from_db

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


def recv_all(socket, host: Optional[str] = None,
             port: Optional[int] = None) -> bytes:
    """Receive all bytes of a message up to total_bytes from a socket."""
    logger.debug(
        f"Thread {threading.current_thread().name} in PID {os.getpid()} is reading socket {socket}")

    try:
        length_prefix = socket.recv(BLOCK_PREFIX_LENGTH_BYTES)
    except (ConnectionError, ConnectionResetError, ConnectionAbortedError,
            ConnectionRefusedError) as e:
        raise e
    finally:
        if host and port:
            asyncio.run(remove_alive_beam_from_db(host, port))

    total_bytes = int.from_bytes(length_prefix, ENDIAN_TYPE, signed=False)
    if total_bytes:
        logger.debug(
            f"Incoming message will have length {total_bytes} B |"
            f" prefix {length_prefix.hex()} | Thread: {threading.current_thread().name}")
    else:
        return b''

    data = b''
    while (data_len := len(data)) < total_bytes:
        # Attempt to read enough bytes to complete the message
        remaining_bytes = total_bytes - data_len
        packet = socket.recv(remaining_bytes)
        if not packet:
            # No more data is being sent; possibly the connection was closed
            logger.debug(
                f"Data is not being sent, closing connection... Remaining {remaining_bytes} Bytes Thread: {threading.current_thread().name} "
                f"| Buffer contained {data}")
            break
        data += packet
    return data



from decentnet.consensus.blockchain_params import BlockchainParams

PORT_DIFFICULTY_CONFIG = {
    8888: BlockchainParams(),
    8889: BlockchainParams(),
    8890: BlockchainParams(),
    8891: BlockchainParams(),
    8892: BlockchainParams()
}



import asyncio
import logging
from functools import lru_cache

import networkx as nx
import sqlalchemy
from sqlalchemy import select
from sqlalchemy.exc import IntegrityError

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.db.base import session_scope
from decentnet.modules.db.models import Edge
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


class FlowNetwork:
    _instance = None  # Class variable to hold the single instance

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(FlowNetwork, cls).__new__(cls)
            # Call the initialization only once.
            cls._instance.__init_once__(*args, **kwargs)
        return cls._instance

    def __init_once__(self, _save_to_db=True):
        self.graph = nx.Graph()
        logger.debug("Rebuilding FlowNetwork from DB")
        if _save_to_db:
            asyncio.run(self.rebuild_flow_network())

    async def rebuild_flow_network(self):
        async with session_scope() as session:
            # Fetch edges asynchronously
            result = await session.execute(select(Edge))
            edges = result.scalars().all()

            # Process each edge
            for edge in edges:
                self.graph.add_edge(edge.pub_key, edge.target, capacity=edge.capacity, flow=0)

            # Log the result
            logger.debug(f"Flow network rebuilt from {len(edges)} edges")

    def add_edge(self, edge_pub_key, target, capacity,
                 _save_to_db: bool = True):
        """
        Adding edge and nodes
        edges will be added to work both ways
        :param edge_pub_key:
        :param target:
        :param capacity:
        :param _save_to_db:
        :return:
        """
        if edge_pub_key == "NOT_KNOWN" or target == "NOT_KNOWN":
            return False

        self.graph.add_edge(edge_pub_key, target, capacity=capacity, flow=0)

        logger.debug(f"Connected edge {edge_pub_key} => {target}")

        if _save_to_db:
            asyncio.run(FlowNetwork.save_edge_to_db(edge_pub_key, target, capacity))

    @lru_cache(maxsize=128)
    def get_path(self, source, sink):
        max_flow = 0
        # Find an augmenting path using Breadth-First Search
        path = nx.shortest_path(self.graph, source=source, target=sink,
                                weight='capacity')
        # Find the minimum capacity on the augmenting path
        capacities = (self.graph[u][v]['capacity'] for u, v in zip(path, path[1:]))

        try:
            min_capacity = min(capacities)
        except ValueError:
            min_capacity = 0
        # Update the flow along the augmenting path
        for u, v in zip(path, path[1:]):
            self.graph[u][v]['capacity'] -= min_capacity
            self.graph[u][v]['flow'] += min_capacity
            self.graph[v][u]['flow'] -= min_capacity
        max_flow += min_capacity

        return path, max_flow

    @classmethod
    async def save_edge_to_db(cls, source_pub_key, target_pub_key, capacity):
        async with session_scope() as session:
            try:
                edge_to = Edge(pub_key=source_pub_key, target=target_pub_key, capacity=capacity)
                session.add(edge_to)
                await session.commit()
                return True
            except (IntegrityError, sqlalchemy.exc.IntegrityError):
                logging.debug(
                    f"Attempted to insert an edge {edge_to.pub_key} => {edge_to.target} that exists")
                await session.rollback()  # Important to roll back the session to a clean state

    def forward_flow(self, path, min_capacity):
        for u, v in zip(path, path[1:]):
            self.graph[u][v]['capacity'] -= min_capacity
            self.graph[u][v]['flow'] += min_capacity
            self.graph[v][u]['flow'] -= min_capacity

    async def bulk_save_db(self):
        try:
            async with session_scope() as session:
                for u, v, data in self.graph.edges(data=True):
                    edge = Edge(source=u, target=v, capacity=data['capacity'], flow=data['flow'])
                    session.add(edge)

                await session.commit()
        except (IntegrityError, sqlalchemy.exc.IntegrityError):
            logging.warning("Attempted to insert an edge that exists")
            await session.rollback()  # Important to roll back the session to a clean state

    def compute_maximum_flow(self, source, sink):
        max_flow = self.get_path(source, sink)

        # Forward flow through the network based on the computed maximum flow
        while True:
            # Find an augmenting path using Breadth-First Search
            path = nx.shortest_path(self.graph, source=source, target=sink,
                                    weight='capacity')
            if not path:
                break

            # Find the minimum capacity on the augmenting path
            min_capacity = min(
                self.graph[u][v]['capacity'] for u, v in zip(path, path[1:]))

            # Forward flow along the augmenting path
            self.forward_flow(path, min_capacity)

        return path, max_flow



import logging
from dataclasses import dataclass

from decentnet.consensus.byte_conversion_constants import ENDIAN_TYPE
from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


@dataclass
class Difficulty:
    t_cost: int
    m_cost: int
    p_cost: int
    n_bits: int
    hash_len_chars: int

    def to_bytes(self) -> bytes:
        return (
                self.t_cost.to_bytes(4, byteorder=ENDIAN_TYPE, signed=False) +
                self.m_cost.to_bytes(4, byteorder=ENDIAN_TYPE, signed=False) +
                self.p_cost.to_bytes(3, byteorder=ENDIAN_TYPE, signed=False) +
                self.n_bits.to_bytes(1, byteorder=ENDIAN_TYPE, signed=False) +
                self.hash_len_chars.to_bytes(1, byteorder=ENDIAN_TYPE, signed=False)
        )

    @classmethod
    def from_bytes(cls, difficulty_bytes: bytes | memoryview):
        t_cost = int.from_bytes(difficulty_bytes[0:4], byteorder=ENDIAN_TYPE,
                                signed=False)
        m_cost = int.from_bytes(difficulty_bytes[4:8], byteorder=ENDIAN_TYPE,
                                signed=False)
        p_cost = int.from_bytes(difficulty_bytes[8:11], byteorder=ENDIAN_TYPE,
                                signed=False)
        n_bits = int.from_bytes(difficulty_bytes[11:12], byteorder=ENDIAN_TYPE,
                                signed=False)
        hash_len_chars = int.from_bytes(difficulty_bytes[12:13], byteorder=ENDIAN_TYPE,
                                        signed=False)

        return cls(t_cost, m_cost, p_cost, n_bits, hash_len_chars)

    def __eq__(self, other):
        """
        Compares all.yaml attributes of the Difficulty instance except for the n_bits attribute.
        """
        if isinstance(other, Difficulty):
            return (self.t_cost, self.m_cost, self.p_cost, self.hash_len_chars) == \
                (other.t_cost, other.m_cost, other.p_cost, other.hash_len_chars)
        return False

    def __repr__(self):
        return f"Difficulty(t_cost={self.t_cost}," \
               f" m_cost={self.m_cost}," \
               f" p_cost={self.p_cost}" \
               f", n_bits={self.n_bits}," \
               f" hash_len_chars={self.hash_len_chars})"

    def __hash__(self):
        return hash(
            (self.t_cost, self.m_cost, self.p_cost, self.n_bits, self.hash_len_chars))

    def __iter__(self):
        yield from [self.t_cost, self.m_cost, self.p_cost, self.n_bits,
                    self.hash_len_chars]

    def __str__(self):
        return f"{self.t_cost}:{self.m_cost}:{self.p_cost}:{self.n_bits}:{self.hash_len_chars}"

    def __post_init__(self):
        if (req_m_cost := 8 * self.p_cost) > self.m_cost:
            logger.debug(
                f"Memory too low increasing from {self.m_cost} to {req_m_cost}")
            self.m_cost = req_m_cost



import asyncio
import logging
from typing import Optional

import cbor2
from ecdsa import SigningKey

from decentnet.consensus.byte_conversion_constants import ENDIAN_TYPE
from decentnet.consensus.dev_constants import RUN_IN_DEBUG, BLOCK_ERROR_DATA_LOG_LEN
from decentnet.modules.blockchain.block import Block
from decentnet.modules.convert.byte_to_base64_slow import bytes_to_base64
from decentnet.modules.cryptography.asymetric import AsymCrypt
from decentnet.modules.key_util.key_manager import KeyManager
from decentnet.modules.logger.log import setup_logger
from decentnet.modules.serializer.serializer import Serializer

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


class Packager:
    @classmethod
    def genesis_package(cls, owner_key_id: int, receiver_key_pub_key: str,
                        block: Block):
        raise NotImplementedError()

    @classmethod
    def check_verified(cls, data: dict, verified: bool):
        """
        If verified is False, this will raise an exception with log
        :param data: Data of block for logging
        :param verified:
        """
        if not verified:
            logger.error(f"Invalid signature, failed to verify data: {data}")
            raise Exception("Invalid signature")

    @classmethod
    def pack(cls, owner_key_id: int, block: Block,
             target_pub_key: Optional[str]) -> bytes:
        o_priv_key, o_pub_key = asyncio.run(KeyManager.retrieve_ssh_key_pair_from_db(owner_key_id))
        block_bytes = block.to_bytes()
        signature = cls.sign_block(owner_key_id, block_bytes)

        data = Serializer.serialize_data(
            o_pub_key, signature,
            block_bytes, target_pub_key)

        if RUN_IN_DEBUG:
            loaded_data = cbor2.loads(data)
            truncated_data = bytes_to_base64(data[:BLOCK_ERROR_DATA_LOG_LEN]) + "..." if len(
                data) > BLOCK_ERROR_DATA_LOG_LEN else loaded_data
            logger.debug("Packed data into %s" % truncated_data)

        return data

    @classmethod
    def sign_block(cls, owner_key_id: int, block_bytes: bytes):
        o_priv_key, _ = asyncio.run(KeyManager.retrieve_ssh_key_pair_from_db(owner_key_id))
        block_bytes = block_bytes
        return AsymCrypt.sign_message(private_key=o_priv_key, data=block_bytes)

    @classmethod
    def add_cmd(cls, unpacked_request: dict, owner_key_id: int, cmd: int) -> dict:
        """
        Adding cmd key to dict
        :param unpacked_request:
        :param owner_key_id:
        :param cmd: This is an int from CMD enum
        :return: Data with added cmd in dictionary
        """
        cpy_unpacked = dict(unpacked_request)
        o_priv_key, o_pub_key = asyncio.run(KeyManager.retrieve_ssh_key_pair_from_db(owner_key_id))
        cmd_bytes = cmd.to_bytes(2, byteorder=ENDIAN_TYPE, signed=False)
        signature = AsymCrypt.sign_message(private_key=o_priv_key, data=cmd_bytes)

        cpy_unpacked["cmd"] = cmd
        cpy_unpacked["csig"] = signature
        cpy_unpacked["cpub"] = o_pub_key
        logger.debug("Added cmd and raw supporting values to packed result: %s" % cpy_unpacked)
        return cpy_unpacked

    @classmethod
    def unpack(cls, serialized_data: bytes, skip_key_verify: bool = False) -> tuple[
        bool, dict, Optional[bool]]:
        """
        Unpacks and verifies block
        :param skip_key_verify:
        :param serialized_data:
        :return: verified and data

        keys:
        "pub"
        "sig"
        "data"
        "target"
        """
        data = Serializer.deserialize_data(serialized_data)
        decoded_pub_key = AsymCrypt.verifying_key_from_string(data["pub"])
        # skip signature verify if encryption is enabled
        if not skip_key_verify:
            verified = AsymCrypt.verify_signature(
                public_key=decoded_pub_key,
                data=data["data"], signature=data["sig"])
        else:
            verified = True

        if data.get("cmd", None) is not None:
            decoded_cpub_key = AsymCrypt.verifying_key_from_string(data["cpub"])
            verified_csig = Packager.verify_csig(data, decoded_cpub_key)
        else:
            verified_csig = None
        return verified, data, verified_csig

    @classmethod
    def verify_csig(cls, data, decoded_cpub_key):
        return AsymCrypt.verify_signature(
            public_key=decoded_cpub_key,
            data=data["cmd"].to_bytes(2, byteorder=ENDIAN_TYPE, signed=False),
            signature=data["csig"])

    @classmethod
    def sign_dict_data(cls, private_key: SigningKey, attribute: str, data: dict):
        """
        This will add signature to data attribute
        :param private_key:
        :param attribute:
        :param data:
        """
        signature = AsymCrypt.sign_message(private_key, data[attribute].encode("ascii"))
        data["sig"] = signature
        return data

    @classmethod
    def verify_dict_data(cls, public_key: str, attribute: str, data: dict):
        """
        Verifies data from a key in dict
        :param public_key:
        :param attribute:
        :param data:
        :return:
        """
        pub_key_raw = AsymCrypt.public_key_from_base64(public_key, False)
        return AsymCrypt.verify_signature(public_key=pub_key_raw, signature=data["sig"],
                                          data=data[attribute].encode("ascii"))



import logging
from wsgiref.simple_server import make_server

import prometheus_client
from prometheus_client.multiprocess import MultiProcessCollector

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.consensus.local_config import registry, PROMETHEUS_PORT
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


class Stats:
    @staticmethod
    def start_prometheus_server():
        logger.debug("Starting prometheus client...")
        MultiProcessCollector(registry)
        app = prometheus_client.make_wsgi_app(registry=registry)
        httpd = make_server('localhost', PROMETHEUS_PORT, app)
        httpd.serve_forever()



"""
Blockchain Parameters
"""
from dataclasses import dataclass

from decentnet.consensus.block_sizing import HASH_LEN
from decentnet.modules.pow.difficulty import Difficulty

SAVE_BLOCKS_TO_DB_DEFAULT = False

@dataclass
class BlockchainParams:
    default_salt = b"Knz3z0&PavluT0m"
    default_salt_len = len(default_salt)
    default_genesis_msg = "CONNECTED"
    seed_difficulty = Difficulty(16, 8, 1, 16, HASH_LEN)
    low_diff = Difficulty(1, 8, 1, 1, HASH_LEN)
    max_hosts = 65536



import rich

from decentnet.consensus.byte_conversion_constants import ENDIAN_TYPE
from decentnet.modules.pow.hashing import hash_func

try:
    from decentnet.modules.pow.computation_fast import int_to_bytes  # noqa
except ImportError as ex:
    rich.print(f"[red]POW Operating in low hashrate mode, more instructions: {ex}[/red]")
    from decentnet.modules.convert.byte_operations import int_to_bytes


class PowUtils:
    @staticmethod
    def get_bit_length(i_hash, nonce, diff) -> int:
        return PowUtils.value_as_int(
            hash_func(int_to_bytes(i_hash.value_as_int() + nonce), diff)).bit_length()

    @staticmethod
    def value_as_int(value) -> int:
        return int.from_bytes(value, ENDIAN_TYPE)



import time
import unittest

from decentnet.consensus.block_sizing import HASH_LEN
from decentnet.modules.dynamic_diff.generate import get_difficulty_parameters
from decentnet.modules.hash_type.hash_type import MemoryHash
from decentnet.modules.pow.difficulty import Difficulty
from decentnet.modules.pow.pow import PoW


@unittest.skip("Not Implemented")
class TestArgon2Parameters(unittest.TestCase):

    def test_base_parameters(self):
        """ Test if the base parameters are correct for index 0 """
        expected = {
            "Memory Cost (KB)": 1024,
            "Time Cost (Iterations)": 1,
            "Parallelism": 2,
            "Zero Bits": 1
        }
        self.assertEqual(get_difficulty_parameters(0), expected)

    def test_incremented_parameters(self):
        """ Test if the parameters increment correctly """
        # Test for index 1
        expected_index_1 = {
            "Memory Cost (KB)": 2048,
            "Time Cost (Iterations)": 2,
            "Parallelism": 2,
            "Zero Bits": 2
        }
        self.assertEqual(get_difficulty_parameters(1), expected_index_1)

        # Test for index 5
        expected_index_5 = {
            "Memory Cost (KB)": 6144,
            "Time Cost (Iterations)": 6,
            "Parallelism": 2,
            "Zero Bits": 6
        }
        self.assertEqual(get_difficulty_parameters(5), expected_index_5)

    def test_high_index(self):
        """ Test if the function handles high indices correctly """
        index = 100
        expected = {
            "Memory Cost (KB)": 1024 + 1024 * index,
            "Time Cost (Iterations)": 1 + index,
            "Parallelism": 2,
            "Zero Bits": 1 + index
        }
        self.assertEqual(get_difficulty_parameters(index), expected)

    def test_linear_time_increment(self):
        """ Test if the time cost increases linearly with the index """
        previous_time_cost = None
        consistent_increment = True
        increment = None

        for index in range(1, 99999999):  # Testing for indices 1 to 10
            tolerance_percentage = 100
            pa = get_difficulty_parameters(index)
            diff = Difficulty(pa["t"], pa["m"], pa["p"], pa["b"], HASH_LEN)
            pw = MemoryHash(diff, bytes(0x86))
            start_time = time.time()
            PoW.compute(pw, diff)
            end_time = time.time()

            current_time_cost = end_time - start_time
            print(
                f"Difficulty: {diff} Index: {index}, Execution Time: {current_time_cost} seconds")

            if previous_time_cost is not None:
                current_increment = current_time_cost - previous_time_cost

                if not increment:
                    increment = current_increment
                else:
                    # Calculate the acceptable range based on the tolerance
                    lower_bound = increment * (1 - tolerance_percentage / increment)
                    upper_bound = increment * (1 + tolerance_percentage / increment)

                    if not (lower_bound <= current_increment <= upper_bound):
                        self.fail(
                            f"Time cost increment is not within the tolerance range at index {index}")

            previous_time_cost = current_time_cost

        self.assertTrue(consistent_increment, "Time cost increment is not linear")


# Running the tests
if __name__ == '__main__':
    unittest.main()



#!/usr/decentmesh/env python3

import os

from setuptools import setup, find_packages

# get key package details from decentmesh/__version__.py
about = {}  # type: ignore
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "decentnet", "__version__.py")) as f:
    exec(f.read(), about)

# load the README file and use it as the long_description for PyPI
with open("README.md", "r") as f:
    readme = f.read()

setup(
    name=about["__title__"],
    description=about["__description__"],
    long_description=readme,
    long_description_content_type="text/markdown",
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    url=about["__url__"],
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.11,<4",
    install_requires=[
        "alembic==1.13.2",
        "argon2-cffi==23.1.0",
        "argon2-cffi-bindings==21.2.0",
        "asn1crypto==1.5.1",
        "Brotli==1.1.0",
        "cffi==1.17.0",
        "click==8.1.7",
        "coincurve==19.0.1",
        "colorama==0.4.6",
        "cryptography==43.0.0",
        "Cython==3.0.11",
        "cytoolz==0.12.3",
        "ecdsa==0.19.0",
        "eciespy==0.4.2",
        "eth-hash==0.5.2",
        "eth-keys==0.4.0",
        "eth-typing==3.5.2",
        "eth-utils==2.3.1",
        "greenlet==3.0.2",
        "lz4==4.3.2",
        "Mako==1.3.0",
        "markdown-it-py==3.0.0",
        "MarkupSafe==2.1.3",
        "mdurl==0.1.2",
        "networkx==3.2.1",
        "pycparser==2.22",
        "pycryptodome==3.20.0",
        "Pygments==2.17.2",
        "pylzma==0.5.0",
        "rich==13.8.0",
        "setuptools==74.0.0",
        "six==1.16.0",
        "SQLAlchemy==2.0.32",
        "toolz==0.12.1",
        "typing_extensions==4.12.2",
        "pymysql==1.1.1",
        "prometheus_client==0.20.0",
        "cbor2~=5.6.4",
        "netifaces==0.11.0",
        "qrcode==7.4.2",
        "sentry-sdk==2.13.0"
    ],
    extras_require={
        "dev": ["black==22.*"],
    },
    license=about["__license__"],
    zip_safe=True,
    entry_points={
        "console_scripts": ["decentmesh=decentnet.main:main"],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    keywords="Decentralized P2P Network",
)
